/* TentacleView.cpp
 * Copyright 1997, Bill Thibault
 */

#include "TentacleView.h"
#include <stdlib.h>
#include <GL/glu.h>
#include "trackball.h"

// configure various option here:
//#define DRAW_WITH_SEPARATE_THREAD
//#define DRAW_SPHERES
//#define LIGHT_PER_TENTACLE
static int	numTentacles = 6;

// number of segments per tentacle
static int     maxLevel = 12;	
// proportions of each tentacle segment
static GLfloat baseSegmentSize[3] = { 2.0, 0.4, 1.0 };
// relative size change for each new segment
static float   scaleAmt = 0.9;

static float curquat[4];
static int mousex, mousey;
static int spinning = 0;
static float lastquat[4];

static GLfloat matTentacleDiffuse[] = { 1.0, 0.0, 0.0, 1.0 };	
static GLfloat matSphereDiffuse[] = { 0.0, 0.0, 0.8, 1.0 };	


static int bend = 0;
static int curl = 0;


static unsigned long	frame = 0;

static GLenum lightID[8] = {
	GL_LIGHT0,
	GL_LIGHT1,
	GL_LIGHT2,
	GL_LIGHT3,
	GL_LIGHT4,
	GL_LIGHT5,
	GL_LIGHT6,
	GL_LIGHT7
	};

// a unit cube centered at the origin
	 GLfloat vx[8][3] = {
		{ -0.5,-0.5,-0.5 },
		{0.5,-0.5,-0.5},
		{0.5,-0.5,0.5},
		{-0.5,-0.5,0.5},
		{-0.5,0.5,-0.5},
		{0.5,0.5,-0.5},
		{0.5,0.5,0.5},
		{-0.5,0.5,0.5}
	};
	 int face[6][4] = {
		{ 0,1,2,3},
		{7,4,0,3},
		{6,7,3,2},
		{5,6,2,1},
		{4,5,1,0},
		{4,7,6,5},
	};
// sphere locations
static const int numSpheres = 8;
	GLfloat sphereLoc[numSpheres][3] = {
		{8,-8,-8},
		{-8,-8,-8},
		{8,8,-8},
		{-8,8,-8},
		{8,-8,8},
		{-8,-8,8},
		{8,8,8},
		{-8,8,8}
		};


TentacleView::TentacleView(BRect rect, char *name) :
	BGLView(rect, name, B_FOLLOW_ALL, 
#ifndef DRAW_WITH_SEPARATE_THREAD
	B_PULSE_NEEDED | 
#endif
	B_WILL_DRAW,
	BGL_DOUBLE | BGL_RGB | BGL_DEPTH )
{
}




TentacleView::~TentacleView ( void )
{
#ifdef DRAW_WITH_SEPARATE_THREAD
	kill_thread ( draw_thread );
#endif
}



void 
TentacleView::AttachedToWindow() {
	GLfloat	matSpecular[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat matAmbient[] = { 0.2, 0.2, 0.2, 1.0 };	
	GLfloat	matShininess[] = { 50.0 };	
	GLfloat	lightPosition[] = { 1.0, 1.0, 1.0, 0.0 };	

	LockGL();
	glClearColor(0.0, 0.0, 0.0, 1.0);  // black background
	glShadeModel ( GL_SMOOTH );
	
	glMaterialfv ( GL_FRONT, GL_SPECULAR, matSpecular );
	glMaterialfv ( GL_FRONT, GL_DIFFUSE, matTentacleDiffuse );
	glMaterialfv ( GL_FRONT, GL_AMBIENT, matAmbient );
	glMaterialfv ( GL_FRONT, GL_SHININESS, matShininess );
	glLightfv ( GL_LIGHT0, GL_POSITION, lightPosition );
	glLightfv ( GL_LIGHT0, GL_AMBIENT, matTentacleDiffuse );
	
	glEnable ( GL_DEPTH_TEST );
 	glEnable ( GL_LIGHTING );
	
#ifndef LIGHT_PER_TENTACLE
	glEnable ( lightID[0] );
#else	
	for ( int i = 0; i < numTentacles; i++ ) {
		glEnable ( lightID[i] );
	}
#endif
	glEnable ( GL_NORMALIZE );
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	BRect rect = Bounds();
	gluPerspective ( 65.0, rect.right / rect.bottom, 1.0 , 100.0 ); 
	glMatrixMode(GL_MODELVIEW);
	UnlockGL();
	
	bend = 0;
	curl = 0;
	
	trackball(curquat, 0.0, 0.0, 0.0, 0.0);

#ifdef DRAW_WITH_SEPARATE_THREAD	
	draw_thread = spawn_thread ( TentacleView::drawer,
									"drawer", 
									/* 2*B_NORMAL_PRIORITY, */
									B_NORMAL_PRIORITY,
									this );
	resume_thread ( draw_thread );
#else
	Window()->SetPulseRate ( 100000 );
#endif
}

void
TentacleView::MouseMoved (  BPoint point, 
					uint32 transit, 
					const BMessage *message ) 
{
	uint32 buttons;
	
	GetMouse ( &point, &buttons );
	if ( buttons & B_PRIMARY_MOUSE_BUTTON ) {
    	if (mousex != point.x || mousey != point.y) {
      		trackball(lastquat,
       	 		(2.0*mousex - Bounds().Width()) / Bounds().Width(),
       	 		(Bounds().Height() - 2.0*mousey) / Bounds().Height(),
        		(2.0*point.x - Bounds().Width()) / Bounds().Width(),
        		(Bounds().Height() - 2.0*point.y) / Bounds().Height());
      		spinning = 1;
    	} else {
      		spinning = 0;
    	}
    }
  
  mousex = point.x;
  mousey = point.y;
}

void
TentacleView::myCube ( float size )
{
	glPushMatrix();
	glScalef ( size, size, size );
	glBegin ( GL_QUADS );
		for ( int i = 0; i < 6; i ++ ) 
			for ( int j = 0; j < 4; j++ ) {
				glNormal3fv ( vx[face[i][j]] );
				glVertex3fv ( vx[face[i][j]] );
			}
	glEnd();
	glPopMatrix();
}

void
TentacleView::mySphere ( float size, int rows, int cols )
{
	myCube ( size );
}


void 
TentacleView::whip ( int id, int level )
{
  GLfloat segsize[3];
  int i;
  static GLfloat lightPos[] = { 1.0, 0.0, 0.0, 1.0 };

  /*
   *  Compute the size of this segment.
   *  As "level" decreases, so does the segment size.
   */

  for ( i=0; i < 3; i++ ) 
    segsize[i] = baseSegmentSize[i] * pow ( scaleAmt, maxLevel-level );

  if ( level > 0 ) {

    /*
     * Move the local coord system for this segment to the pivot point,
     * at the "right" end of the PREVIOUS segment.
     * This is in terms of the PREVIOUS segment's coordinate system.
     * That segment was bigger by a factor of scaleAmt, so
     * divide by that here.
     */

    glTranslatef ( segsize[0]/(2.0 * scaleAmt), 0, 0 );

    /*
     * Rotate by the current curl amount. 
     */

    glRotatef ((GLfloat) curl, 1.0, 0.0, 0.0);

    /*
     * Now rotate by the current bend amount. 
     */

    glRotatef ((GLfloat) bend, 0.0, 0.0, 1.0);

    /*
     * In preparation for drawing, move the local coord system origin
     * to the center of the segment to be drawn. (since we're using
     * glutWireCube() which draws a cube centered at the origin.
     */
    glTranslatef ( segsize[0]/2.0, 0, 0 );

    /*
     * Draw the segment as a scaled cube. 
     */

    glPushMatrix();
    glScalef ( segsize[0], segsize[1], segsize[2] );
   	glMaterialfv ( GL_FRONT, GL_DIFFUSE, matTentacleDiffuse );

    myCube (1.0);
    glPopMatrix();
    
    /*
     * Recurse for the rest of the whip.
     */

    whip ( id, level - 1 );
  }
  else {

    /* 
     * put the light at the very end
     */
    lightPos[0] = segsize[0];
    	glLightfv ( lightID[id], GL_POSITION, lightPos );


  }
}


void 
TentacleView::display(void)
{
   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glMaterialfv ( GL_FRONT, GL_DIFFUSE, matTentacleDiffuse );
   	mySphere ( 2.0, 10, 10 );
   	
	for ( int i = 0; i < numTentacles; i++ ) {
		glPushMatrix();

		// glTranslatef ( -baseSegmentSize[0]/2.0, 0, 0 );
		
		switch ( i ) {
			case 0:
				break;
			case 1:
				glRotatef ( 180, 0, 0, 1 );
				break;
			case 2:
				glRotatef ( 90, 0, 0, 1 );
				break;
			case 3:
				glRotatef ( -90, 0, 0, 1 );
				break;
			case 4:
				glRotatef ( -90, 0, 1, 0 );
				break;
			case 5:
				glRotatef ( 90, 0, 1, 0 );
				break;
			case 6:
				glRotatef ( -45, 0, 1, 0 );
				break;
			case 7:
				glRotatef ( 45, 0, 1, 0 );
				break;
			default:
				break;
		}
		
		whip ( i, maxLevel );
	
		glPopMatrix();
	}

#ifdef DRAW_SPHERES
	// draw its nest
   	glMaterialfv ( GL_FRONT, GL_DIFFUSE, matSphereDiffuse );
	for ( int i = 0; i < numSpheres; i++ ) {
		glPushMatrix();
		glTranslatef ( sphereLoc[i][0],
						sphereLoc[i][1],
						sphereLoc[i][2] );
		mySphere( 4.0, 10, 10 );
		glPopMatrix();
	}
#endif

}

void 
TentacleView::FrameResized ( float width, float height )
{
	inherited::FrameResized ( width, height );
	LockGL();
  	glViewport(0, 0, width, height);
  	UnlockGL();
}

void TentacleView::Draw(BRect) 
{
	float m[4][4];
	LockGL();
	
	glPushMatrix();
	
  		if ( spinning ) {
    		add_quats ( lastquat, curquat, curquat );
  		}
  		build_rotmatrix(m, curquat);
  		glMatrixMode(GL_MODELVIEW);
  		glLoadIdentity();
  		glTranslatef(0, 0, -30);
  		glMultMatrixf(&(m[0][0]));

		display();
		
	glPopMatrix();
	
	glFlush();
	SwapBuffers();
	UnlockGL();
}



void
TentacleView::Pulse ( void )
{
	static unsigned long	i;
	
	GLfloat	center = 1.0;
	
	frame++;
	bend = 30 * sin ( 7 * ((2*PI)/360.0 ) * (frame % 360) );
	curl = 20  * sin ( 3 * ((2*PI)/360.0 ) * (frame % 360 ) );
	Draw(Bounds());
}

long
TentacleView::drawer ( void *param )
{
	BGLView *v = (BGLView *)param;
	while ( 1 ) {
		v->Window()->Lock();
		v->Pulse ();
		v->Window()->Unlock();
		
		snooze ( 100 * 1000.0 );
	}
	return 1;
}